expit<-function(x) 1/(1+exp(-x))


### function to assign value of "x" to the j^th position of a vector beta

changeXj<-function(beta,j,x){
          betanew=beta
          betanew[j]=x
          betanew
        }


#########################################################################################################
### Function to compute weights for deriving rules that minimizing weighted
### sum of 0-1 Loss
######## Input
####     data: with the following variables
####          -------- yy: binary disease outcome
####          -------- TT: binary treatment indicator
####          -------- xi: for i in 1:M with M a positive integer
####          -------- weights: sampling weights for individuals, default to 1 if missing
####     cost: treatment/disease burden ratio delta
####
####### Output
####### dataset that include all variables in "data", additional variables are
####### the four types of weights for computing the sum of weighted 0-1 loss
####### Details about those four types of weights can be found in Huang and Fong 2014 Biometrics  ###
##########################################################################################################


calweights<-function(dat,cost){

       xx=xx.tt=xx.tt0=xx.tt1=NULL
      for (j in 1:m.x){
       xj<-paste('x',j,sep='')
       xx<-cbind(xx,dat[,xj])
      }
      xx<-cbind(xx,dat$TT)
      for (j in 1:m.x){
       xj<-paste('x',j,sep='')
       xx<-cbind(xx,(dat$TT)*dat[,xj])
      }
    if (is.null(dat$weights)) {
        weights=rep(1,nrow(dat))
    } else {
        weights=dat$weights
    }

    fit0=cv.glmnet(xx,dat$yy,weights=weights,family="binomial")

    X.f<-cbind(1,xx[,1:m.x])

    n0=sum(dat$TT==0);n1=sum(dat$TT==1)

    #### weights consider disease outcome as 0,1
     dat$W1.f=(-1)*(dat$yy*dat$TT/n1-dat$yy*(1-dat$TT)/n0+cost/(n0+n1))     ## type-I
     dat$W2.f=(-1)*((1-dat$yy)*(1-dat$TT)/n0-(1-dat$yy)*dat$TT/n1+cost/(n0+n1))  ### type-II
     dat$W3.f=dat$W1.f/2+dat$W2.f/2                                         ## type-III

     oo0<-1:(length(coef(fit0))/2)
     oo1<-(length(coef(fit0))/2+1):(length(coef(fit0)))

    ################

     DD.s<-expit(X.f%*%coef(fit0)[oo0])+expit(X.f%*%coef(fit0)[oo0]+X.f%*%coef(fit0)[oo1])

     dat$W4.f=(-1)*(dat$yy*dat$TT/n1-dat$yy*(1-dat$TT)/n0+cost/(n1+n0)+(1-2*dat$TT)/(n1+n0)*DD.s)
     dat$W4.f=as.numeric(dat$W4.f)                                       ## type-IV

     ################## incorporating sampling weights ##################

     dat$W1.f<-dat$W1.f*weights
     dat$W2.f<-dat$W2.f*weights
     dat$W3.f<-dat$W3.f*weights
     dat$W4.f<-dat$W4.f*weights



     dat$W1.f=dat$W1.f/mean(abs(dat$W1.f))
     dat$W2.f=dat$W2.f/mean(abs(dat$W2.f))
     dat$W3.f=dat$W3.f/mean(abs(dat$W3.f))
     dat$W4.f=dat$W4.f/mean(abs(dat$W4.f))

     return(dat)
}


######## Function to compute theta (total disease and treatment cost) based on samples from a randomized trial
######## Input
####     Y: binary disease outcome
####     TT: binary treatment indicator
####     score: predicted score for treatment selection, i.e. f(X) such that treatment is recommended if f(X)>0
####     weights: sampling weights for individuals, default to 1 if missing
####     cost: treatment/disease burden ratio delta
####
####### Output
####### theta (total disease and treatment cost)

caltheta.weights<-function(Y,TT,score,weights=NULL,cost=0){
 if (sum(!is.na(score))==0) out=NA
 else {
 if (is.null(weights)) weights=rep(1,length(Y))
 out=sum(Y*(score>0)*TT*weights,na.rm=T)/sum(TT*weights,na.rm=T)+sum(Y*(score<=0)*(1-TT)*weights,na.rm=T)/sum((1-TT)*weights,na.rm=T)+
 sum((score>0)*weights,na.rm=T)/sum(weights,na.rm=T)*cost
 }
 return(out)
}

######## Function to compute percent to be treated based on samples from a randomized trial
######## Input
####     score: predicted score for treatment selection, i.e. f(X) such that treatment is recommended if f(X)>0
####     weights: sampling weights for individuals, default to 1 if missing
####
####### Output
#######  percent to be treated based on the treatment-selection rule


caltreat.weights<-function(score,weights){
 if (sum(!is.na(score))==0) out=NA
 else {
 out=sum((score>0)*weights,na.rm=T)/sum(weights,na.rm=T)
 }
 return(out)
}


################################################################################################################
######## The smooth function "S(u)" in Huang 2015 that approximates the hinge function "h1(u)"  #############
######## Input:
########  x: the input vector
########  epsilon: the parameter in S(u)
######## Output:
########  S(x)
################################################################################################################
S4<-function (x, epsilon){
0.5 - epsilon * (log(1 + exp(0.5/epsilon)) - ifelse((0.5 - x)/epsilon >
    700, (0.5 - x)/epsilon, log(1 + exp((0.5 - x)/epsilon))))}

######## The function to compute derivative of "S(u)" ##########################################
S4.deriv<-function (x, epsilon)
-1/(1 + exp((x - 0.5)/epsilon))



rauc.dca.approx<-function (beta, X, Z,linear.coef.in.dca, epsilon, Wabs)
{
    eta = Z*(X %*% beta)
    sum(Wabs*(S4(eta, epsilon) - eta * linear.coef.in.dca))
}

rauc.dca.approx.gr<-function (beta, X,Z, linear.coef.in.dca, epsilon,Wabs,index=NULL)
{
    eta = Z*(X %*% beta)
    out=colSums(drop(S4.deriv(eta, epsilon) - linear.coef.in.dca) *
        Z*X*Wabs)
    if (!is.null(index)) out=out[index]
    return(out)
}

######################### Function to compute the RAMP Loss ############
#### Input:
#### eta: a vector of paired difference of linear combinations
#### s:	 absolute value of the slope parameter
#### loss: 	a boolean. If TRUE, return loss function i.e. 1 - RAUC. If FALSE, rturn RAUC. Default to TRUE, because we minimize loss.
#### Output:
#### A vector of sames size as eta
###########################################################################


ramp.f<-function (eta, s, loss = TRUE)
{
    if (s == 0) {
        out = ifelse(eta < 0, 1, 0)
    }
    else {
        out = ifelse(eta > -0.5 * s & eta < 0.5 * s, 0.5 - 1/s *
            eta, 0)
        out[eta <= -0.5 * s] = 1
    }
    if (loss)
        drop(out)
    else drop(1 - out)
}


######### function for finding marker combination that minimizes penalized weighted sum of ramp loss:
####### \sum_{i=1}^n W_ih(X_i^T beta))/n+\lambda*||\beta||^2+\lambdaS*\sum_{i=1}|\beta_i|
##### Input:
######
##### formula:   an object of class "formula" (or one that can be coerced to that class): a symbolic description of the model to be fitted
##### dat:       training data (a data frame)
##### W:         vector of weights
##### s:         tuning parameter for the ramp loss, default to 1
##### lambda:    tuning parameter for L2 norm, default to 0
##### lambdaS:   tuning parameter for L1 norm
##### kernel:    kernel function between covariate, one of 'linear' ('l'), 'rbf' ('r' or 'e'), 'ibs' ('i')
##### eta0.init: initial value for sgn(w_i)f(x_i)
##### maxit:     max number of iteration for DCA
##### tol:       tolerance for convergence of DCA
##### verbose:   logical indicator about whether to output each iterative step
##### cost:      treatment/disease burden ratio delta
##### beta.init: initial value for beta for linear kernel
#####
##### Output: a result object containing following components
#####
##### converge:           indicator of whether the DCA algorithm converged, 0 for converged
##### iteration:          number of iteration
##### b0:                 estimate of intercept
##### coefficients:       estimate of coefficient for covariate for linear kernel, NULL if nonlinear kernel
##### W:                  weight
##### X:                  model matrix
##### alpha.pred:         Lagrange multipliers, will be NA in SRAMP-L1 estimator
##### kernel:             kernel function
##### para:               additional tuning parameter for rbf or ibs kernel, see getK for details, equals NULL in SRAMP-L1 estimator
##### formula:            an object of class "formula" (or one that can be coerced to that class): a symbolic description of the model to be fitted
##### lambda:             tuning parameter for L2 penalty, default to 0
##### lambdaS:            tuning parameter for L1 penalty, default to 0

SRAMP_lasso.W1=function (formula, dat, W, s = 1, lambda = 0, lambdaS=0, eta0.init = NULL,
    maxit = 100, tol = 1e-03, verbose = F, cost=0, beta.init, epsilon=0.001)
{

    method='Nelder-Mead'
    kernel='l'
    para=NULL

    time0 = Sys.time()
    tmp = model.frame(formula, dat)

    X = model.matrix(formula, tmp)[, -1, drop = F]  ###
    n = nrow(X)  ###
    TT = dat$TT


    ### get rid of points where W=0
    oo<-W!=0
    W<-W[oo]
    X<-as.matrix(X[oo,])


    Z<-sign(W)
    Wabs<-abs(W)   # absolute value for W
    Wm<-diag(Wabs)
    Zm<-diag(Z)
    kernel = substr(kernel, 1, 1)
    if (kernel != "l" & is.null(para))
        stop("kernel parameter not set")
    K = getK(X, kernel, para)/lambda
    if (!is.double(K))
        K <- as.double(K)
    date()


    if (is.null(eta0.init)) {

        if (kernel == "l"){
          eta.init = Z * (cbind(1,X) %*% beta.init)
        } else
        eta.init=rep(0,sum(oo))
    } else eta.init=eta0.init

    val.init = sum(Wabs*ramp.f(eta.init,s))+lambda*sum(beta.init[-1]^2)/2+lambdaS*sum(abs(beta.init[-1]))/2

      XX<-cbind(1,X)
      KK<-diag(ncol(X))

    iter = 0
    converge = 0
    beta.new=beta.init
    while (TRUE) {
        iter = iter + 1
        if (iter > maxit) {
            converge = 1
            break
        }
        dh2.deta = ifelse(eta.init < -0.5/s, -s, 0)
        linear.coef.in.dca<-dh2.deta

       iter.inner=0
       val.inner.init=rauc.dca.approx(beta.new, XX, Z,linear.coef.in.dca, epsilon, Wabs)+
       lambda*sum(beta.new[-1]^2)/2+lambdaS*sum(abs(beta.new[-1]))/2

       ### implementation of the coordinate descent algorithm

       while (TRUE) {
        iter.inner = iter.inner + 1
        if (iter.inner > 50) {
            break
        }
       beta.initS1<-optimize(f = function(x) rauc.dca.approx(c(x,beta.new[-1]),
                  XX, Z, linear.coef.in.dca, epsilon, Wabs), interval=c(-10,10),tol=1e-6)

        beta.new[1]=beta.initS1$minimum

        for (j in 2:length(beta.new)){

        fit.zero<-rauc.dca.approx(changeXj(beta.new,j,0),
                  XX, Z, linear.coef.in.dca, epsilon, Wabs) + 0.5 *
                  lambda * c(crossprod(changeXj(beta.new,j,0)[-1],KK) %*% changeXj(beta.new,j,0)[-1])

              fit.pos<-optimize(f = function(x) rauc.dca.approx(changeXj(beta.new,j,x),
                  XX, Z, linear.coef.in.dca, epsilon, Wabs) + 0.5 *
                  lambda * c(crossprod(changeXj(beta.new,j,x)[-1],KK) %*% changeXj(beta.new,j,x)[-1])+
                  0.5*lambdaS*(sum(abs(beta.new[-c(1,j)]))+x),interval=c(-10,10),tol=1e-6)

         fit.neg<-optimize(f = function(x) rauc.dca.approx(changeXj(beta.new,j,x),
                  XX, Z, linear.coef.in.dca, epsilon, Wabs) + 0.5 *
                  lambda * c(crossprod(changeXj(beta.new,j,x)[-1],KK) %*% changeXj(beta.new,j,x)[-1])+
                  0.5*lambdaS*(sum(abs(beta.new[-c(1,j)]))-x), interval=c(-10,10),tol=1e-6)

         if (fit.pos$minimum>=0 & fit.pos$objective<=fit.zero) {
                beta.new[j]=fit.pos$minimum
               }  else if (fit.neg$minimum<=0 & fit.neg$objective<=fit.zero){
                beta.new[j]=fit.neg$minimum
               } else {
               beta.new[j]=0
               }
             }
              eta.new=Z*(XX%*%beta.new)
              val.inner.new = rauc.dca.approx(beta.new, XX, Z,linear.coef.in.dca, epsilon, Wabs)+
                lambda*c(crossprod(beta.new[-1], KK) %*% beta.new[-1])/2+lambdaS*sum(abs(beta.new[-1]))/2

              if (verbose) {
                    cat(", val.inner.new ", val.inner.new)
              }
                delta=abs(val.inner.new-val.inner.init)
                if (delta < 0.01) {
                    break
                }
                else {
                    eta.init = eta.new
                    val.inner.init=val.inner.new
                }

            }
              eta.new=Z*(XX%*%beta.new)

              val.new = sum (Wabs * ramp.f(eta.new,s))

              pen1 = lambda*c(crossprod(beta.new[-1], KK) %*% beta.new[-1])/2
              pen2 = lambdaS*sum(abs(beta.new[-1]))/2
              if (verbose)
                cat(", saturated ", signif(mean(eta.new > 0.5/s |
                    eta.new < -0.5/s), 3))
            if (kernel == "l") {

                if (verbose)
                  cat(", beta[1] ", signif(beta.new[1], 4), ", beta[-1] ",
                   signif(beta.new[-1], 4))
                val.new = val.new + pen1 + pen2
                if (verbose) {
                    cat(", loss ", val.new, ", pen1 ", pen1, "pen2", pen2)
                    cat(", val.new ", val.new)
                }
                 delta=abs(val.new-val.init)
                if (delta < tol) {
                    converge = 0
                    break
                }
                else {
                    beta.init = beta.new
                    eta.init = eta.new
                    val.init=val.new
                }
          }
          if (verbose)
          cat(iter,"\n")
      }

        #####################

    if (verbose)
        cat("\n")

         result = list(converge = converge, iteration = iter, b0=NA,coefficients = NA, W=W,X=X,
        kernel = kernel, para = para, formula = formula,alpha.pred=NA,lambda=lambda,lambdaS=lambdaS)
    class(result) = c("rauc", class(result))
    if (kernel == "l") {
        result$coefficients = beta.new[-1]
        result$b0=beta.new[1]

    }
    return(result)
}

############################################################
#### Function to predict the marker score f(X) such that treatment is recommended if f(X)>0
#### Input:
#### fit: fitted object from SRAMP_lasso
#### newdata: test data
predict.rtw2<-function (fit, newdata, case.percentage = NULL)
{
    if (inherits(fit,'try-error'))
        return(rep(NA, nrow(newdata)))
    if (is.null(fit$alpha.pred))
        return(rep(NA, nrow(newdata)))
    ff = fit$formula
    tt <- terms(ff)
    ff1 = reformulate(attr(delete.response(tt), "term.labels"))
    X.pred = model.matrix(ff1, newdata, na.action=NULL)[, -1, drop = F]
    if (fit$kernel == "l") {
        linear.combination = fit$b0 + X.pred %*% coef(fit)
    }
    else {
        K = getK(X.pred, fit$kernel, fit$para, X2 = fit$X)
        linear.combination = fit$b0+crossprod(t(K),fit$alpha.pred*sign(fit$W))/fit$lambda

    }
    res = drop(linear.combination)
    if (!is.null(case.percentage)) {
        threshold = quantile(res, 1 - case.percentage)
        attr(res, "Class") = as.numeric(res > threshold)
    }
    res
}
